/*-
 * Copyright (C) @BABOLO  2009 May 07 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bsdfig.h,v 1.24 2011/11/27 16:28:09 babolo Exp $
 */

#ifndef _SYS_PARAM_H_
# include <sys/param.h>
#endif
#ifndef _SYS_TYPES_H_
# include <sys/types.h>
#endif
#ifndef _STDIO_H_
# include <stdio.h>
#endif
#ifndef _ALIAS_H_
# include <alias.h>
#endif
#ifndef _MIFE_H_
#include <mife.h>
#endif
#ifndef _MULTILAR_H_
#include <multilar.h>
#endif

#ifndef _BSDFIG_H_
#define _BSDFIG_H_
__BEGIN_DECLS

#define BSDFIG_IPFW_DISABLES_MAX_S 6  /* binary incompatible when change */
#define BSDFIG_IPFW_TABLES_MAX_S   8  /* binary incompatible when change */
#define BSDFIG_IPFW_OTHERS_MAX_S   16 /* binary incompatible when change */
#define BSDFIG_IPFW_DISABLES_MAX   (1 << BSDFIG_IPFW_DISABLES_MAX_S)
#define BSDFIG_IPFW_TABLES_MAX     (1 << BSDFIG_IPFW_TABLES_MAX_S)
#define BSDFIG_IPFW_OTHERS_MAX     (1 << BSDFIG_IPFW_OTHERS_MAX_S)
#define BSDFIG_BASELEN             64
#define BSDFIG_BYTELEN             8
#define BSDFIG_RULENHEAD           ((sizeof(struct ip_fw) / sizeof(bsdfig_syllable)) - 1)
#define BSDFIG_ARGTY(X)            (1ULL << (X))

typedef u_int32_t bsdfig_syllable;

typedef struct bsdfig_syntax {
    u_int64_t magik0;
    u_char    magik[8];
    size_t last_opcode;
    struct bsdfig_property {
        union {
            struct {
               /********
                * 0000000000111111111122222222223333333333444444444455555555556666
                * 0123456789012345678901234567890123456789012345678901234567890123
                * llllllLLLLLLMMMMRRRRRRRRPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP--------
                */
                u_int64_t l: 6             // llllll   minimal length of this operation
                ,         L: 6             // LLLLLL   maximal length of this operation
                ,         M: 4             // MMMM     length modificators
#                   define BSDFIG_LMDLM  1   //   
#                   define BSDFIG_LMDLI  2   //   1 + 8n
                             /********
                              *     (  Ap..Aa
                              *   nX/Ms).  Ap..Aa    
                              *  ,     nX 
                              */
                ,         R: 8             // RRRRRRRR rule role
#                   define BSDFIG_ROLpl 0x01 //   
#                   define BSDFIG_ROLAp 0x02 // PROB()
#                   define BSDFIG_ROLAs 0x04 // PROBE_STATE
#                   define BSDFIG_ROLAc 0x06 //  
#                   define BSDFIG_ROLAk 0x08 // KEEP_STATE, LIMIT()
#                   define BSDFIG_ROLAl 0x10 // LOG()
#                   define BSDFIG_ROLAq 0x12 // ALTQ()
#                   define BSDFIG_ROLAt 0x14 // TAG()
#                   define BSDFIG_ROLAa 0x1C //  actions
#                   define BSDFIG_ROLAn 0x1E // Never this level
#                   define BSDFIG_ROLAA 0x10 // action mask
#                   define BSDFIG_ROLAm 0x1E // mask
                ,         P:32             // PP...PPP operation representation
#                   define BSDFIG_A1NUM  0   // arg1 is N of words in d[]
#                   define BSDFIG_A1BIT  1   // arg1 is N of bits after d[0]
#                   define BSDFIG_A163K  2   // arg1 full range
#                   define BSDFIG_A1255  3   // arg1 0..255
#                   define BSDFIG_A1256  4   // arg1 2..256
#                   define BSDFIG_A1127  5   // table num
#                   define BSDFIG_A1FIB  5   // arg1 == FIB N
#                   define BSDFIG_A1_3   6   // arg1 0..3
#                   define BSDFIG_A1_BB  7   // arg1 set_mask reset_mask
#                   define BSDFIG_PAVAR  8   // arg1 or H pairs in words
#                   define BSDFIG_WFULL  9   // d[0] full range
#                   define BSDFIG_FLOAT 11   // d[0] float
#                   define BSDFIG_WIP4N 12   // d[0] IP4 net
#                   define BSDFIG_WIP4H 13   // d[0] IP4 host
#                   define BSDFIG_IFA4  14   // IP4 iface
#                   define BSDFIG_WIP6N 15   // d[0...] IP6
#                   define BSDFIG_ETHER 16   // d[0...] MACdst MACsrc maskdst masksrc
#                   define BSDFIG_SOCK  17   // d[0...] sockaddr
#                   define BSDFIG_LIMIT 18   // LIMIT
#                   define BSDFIG_PAIRS 19   // H pairs in words
#                   define BSDFIG_WORDS 21   // d[0...] DDDDDD ...
#                   define BSDFIG_TEXT0 22   // d[0...] text
#                   define BSDFIG_TEXT1 23   // d[1...] text
#                   define BSDFIG_BITS0 24   // d[0...] bits
#                   define BSDFIG_BITS1 25   // d[1...] bits
#                   define BSDFIG_IP4MS 26   // IP4&IP4mask ...
#                   define BSDFIG_IP6MS 27   // IP6&IP6mask ...
#                   define BSDFIG_ERROR 31   // Error
                ;
            };
            u_int64_t     p;
        };
       /********
        *   ,  ,  NULL,    
        * 1 -    ,      
        */
        const bsdfig_syllable *m;
        char                  *opn;
        u_int8_t               code;
    } p[];
} bsdfig_syntax;

typedef struct bsdfig_ipfwrules {
    socklen_t   size;
#ifndef _IPFW2_H
    void        *rules[];
#else
    struct ip_fw rules[];
#endif
} bsdfig_ipfwrules;

#if __FreeBSD_version >= 700000
typedef struct bsdfig_ipfwnats {
    socklen_t   size;
    int         nat_cnt;
#   ifndef _IPFW2_H
        void          *nats[];
#   else
        struct cfg_nat nats[];
#   endif
} bsdfig_ipfwnats;

typedef struct bsdfig_ipfwnatl {
    socklen_t   size;
#   ifndef _IPFW2_H
        void          *natl[];
#   else
        struct {
            int     natnum;
            uint8_t natlog[LIBALIAS_BUF_SIZE];
        }              natl[];
#   endif
} bsdfig_ipfwnatl;
#endif

typedef struct bsdfig_ipfwtables {
    socklen_t   tabn;
#ifndef _IPFW2_H
    void       *tables[];
#else
    ipfw_table *tables[];
#endif
} bsdfig_ipfwtables;

typedef struct bsdfig_dummynets {
    socklen_t size;
    void     *dummies[];
} bsdfig_dummynets;

#ifndef _IPFW2_H
typedef void ip_fw;
typedef void dn_pipe;
typedef void dn_flow_set;
#else
typedef struct ip_fw ip_fw;
typedef struct dn_pipe dn_pipe;
typedef struct dn_flow_set dn_flow_set;
#endif

typedef struct bsdfig_allconf {
    u_int32_t          flags;
    u_int16_t          fibn;
    u_int16_t          tabn;
    u_int64_t          my_pipes     [BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BASELEN];
    u_int64_t          my_queues    [BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BASELEN];
    u_int64_t          my_ipfwrules [BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BASELEN];
    u_int64_t          my_ipfwnats  [BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BASELEN];
    u_int64_t          my_ipfwtables[BSDFIG_IPFW_TABLES_MAX / BSDFIG_BASELEN]; /*  256  */
    u_int64_t          my_disables  [1];
    u_int64_t          my_fibs      [1];
    void              *my_routes;
    void              *my_ifs;
    u_int64_t          set_disable  [1];
    mular_descriptor  *dummynets;
    mular_descriptor  *ipfwrules;
    mular_descriptor  *ipfwnats;
    mular_descriptor **ipfwtables;
    mular_descriptor **routes;
    mular_descriptor  *ifs;
} bsdfig_allconf;

typedef struct bsdfig_conf {
    u_int32_t            flags;
#       define BSDFIG_EXLEW 0x00000001 /*      1                 */
/*  ,               */
#       define BSDFIG_EXLEI 0x00000002 /*      2                 */
/*   ,   ,  ,                   */
#       define BSDFIG_EXLEE 0x00000004 /*      3                 */
/*    ,    ,                            */
#       define BSDFIG_EXLES 0x00000008 /*      4                 */
/*    , , ,       ,      */
#       define BSDFIG_EXLEM 0x0000000F /*   "     ..."   */
#       define BSDFIG_VDUMP 0x00000010 /*    ipfw                     */
#       define BSDFIG_LPIPE 0x00000020 /*           */
#       define BSDFIG_LQUEU 0x00000040 /*       */
#       define BSDFIG_LSSET 0x00000080 /*        */
#       define BSDFIG_LRUFW 0x00000100 /*         */
#       define BSDFIG_LNATS 0x00000200 /*      NAT    */
#       define BSDFIG_LFTAB 0x00000400 /*         */
#       define BSDFIG_PPIPE 0x00004000 /*                                    */
#       define BSDFIG_DPIPE 0x00008000 /*                                  */
#       define BSDFIG_QUEUE 0x00010000 /*                                */
#       define BSDFIG_DQUEU 0x00020000 /*                              */
#       define BSDFIG_PRSET 0x00040000 /*                       */
#       define BSDFIG_PRUFW 0x00080000 /*                             */
#       define BSDFIG_PRDFW 0x00100000 /*                            */
#       define BSDFIG_CNATS 0x00200000 /*   NAT                                */
#       define BSDFIG_DNATS 0x00400000 /*   NAT                              */
#       define BSDFIG_FTABL 0x00800000 /*                                        */
#       define BSDFIG_0CONF 0x10000000 /*                      */
#       define BSDFIG_MYCNF 0x20000000 /*                          */
#       define BSDFIG_READC 0x40000000 /* read configs                                        */
#       define BSDFIG_KTEST 0x80000000 /*                     */
    u_int16_t            fibn;
    u_int16_t            tabn;
    const bsdfig_syntax *syntax;
    FILE                *out;
    bsdfig_allconf      *conf;
} bsdfig_conf;

extern const       bsdfig_syntax bsdfig_s;
void               bsdfig_syntaxdump       __P((bsdfig_conf *cnf));
bsdfig_conf       *bsdfig_init             __P((u_int32_t flags, FILE *out, bsdfig_conf *slug));
#if __FreeBSD_version >= 700000
bsdfig_ipfwnats   *bsdfig_read_ipfwnats    __P((bsdfig_conf *cnf));
bsdfig_ipfwnatl   *bsdfig_read_ipfwnatl    __P((bsdfig_conf *cnf));
#endif
bsdfig_dummynets  *bsdfig_read_dummynets   __P((bsdfig_conf *cnf));
bsdfig_ipfwrules  *bsdfig_read_ipfwrules   __P((bsdfig_conf *cnf));
bsdfig_ipfwtables *bsdfig_read_ipfwtables  __P((bsdfig_conf *cnf));
#if __FreeBSD_version >= 700000
int                bsdfig_ckout_nats       __P((bsdfig_conf *cnf, bsdfig_ipfwnats   *nats));
int                bsdfig_ckout_natl       __P((bsdfig_conf *cnf, bsdfig_ipfwnatl   *natl));
#endif
int                bsdfig_ckout_pipes      __P((bsdfig_conf *cnf, bsdfig_dummynets  *dummies));
int                bsdfig_ckout_queues     __P((bsdfig_conf *cnf, bsdfig_dummynets  *dummies));
int                bsdfig_ckout_ipfwrules  __P((bsdfig_conf *cnf, bsdfig_ipfwrules  *rules));
int                bsdfig_out_ipfwset      __P((bsdfig_conf *cnf, bsdfig_ipfwrules  *rules));
int                bsdfig_ckout_ipfwtables __P((bsdfig_conf *cnf, bsdfig_ipfwtables *tables));
int                bsdfig_confread         __P((bsdfig_conf *cnf, mife_descriptor   *in));
int                bsdfig_pipelist         __P((bsdfig_conf *cnf, u_int64_t *list));
int                bsdfig_queuelist        __P((bsdfig_conf *cnf, u_int64_t *list));
int                bsdfig_disablelist      __P((bsdfig_conf *cnf, u_int64_t *list));
int                bsdfig_ipfwrulelist     __P((bsdfig_conf *cnf, u_int64_t *list));
#if __FreeBSD_version >= 700000
int                bsdfig_ipfwnatslist     __P((bsdfig_conf *cnf, u_int64_t *list));
#endif
int                bsdfig_ipfwtablelist    __P((bsdfig_conf *cnf, u_int64_t *list));
int                bsdfig_out_pipe         __P((bsdfig_conf *cnf, dn_pipe     *pipe));
int                bsdfig_out_queue        __P((bsdfig_conf *cnf, dn_flow_set *queue));
int                bsdfig_ckout_ipfwrule   __P((bsdfig_conf *cnf, ip_fw       *rule));

__END_DECLS
#endif /* !_BSDFIG_H_ */
